"""
    Fastberry Main (Core-Methods)
"""

import strawberry
from starlette.middleware.base import BaseHTTPMiddleware as BaseMiddleware
from strawberry.extensions import Extension as BaseExtension
from strawberry.permission import BasePermission

from . import config

# from .config import Settings
from .config import Settings as Fastberry
from .config.extras.gql import GQL

# from .graphql import BaseType
from .schema import Schema
from .types import Model, SQLBase, SQLFilters
from .utils.coro import coro
from .utils.to_camel_case import to_camel_case

# Custom Field Types
Text = Model.text
JSON = Model.json
ID = strawberry.ID



"""
    SQLAlchemy + Alembic
"""

import os
import shutil
import pathlib
import click

try:
    from alembic import command
    from alembic.config import Config

    ALEMBIC_CONFIG = Config("alembic.ini")

except ImportError:
    ALEMBIC_CONFIG = None


@click.group()
def db():
    """Database Migrations Commands"""


@db.command()
@click.option("-m", "--message", help="Migration message.", type=str, default=None)
def make_migrations(message):
    """Database Make-Migrations."""
    command.revision(ALEMBIC_CONFIG, message=message, autogenerate=True)


@db.command()
def migrate():
    """Database Migrate."""
    command.upgrade(ALEMBIC_CONFIG, "head")


@db.command()
@click.option("-m", "--message", help="Migration message.", type=str, default=None)
def auto_migrate(message):
    """Database Make-Migrations & Migrate."""
    # Make-Migrations
    command.revision(ALEMBIC_CONFIG, message=message, autogenerate=True)
    # Migrate
    command.upgrade(ALEMBIC_CONFIG, "head")


@db.command()
@click.argument("revision", type=str)
def upgrade(revision):
    """Database Migrate (Upgrade)."""
    command.upgrade(ALEMBIC_CONFIG, revision)


@db.command()
@click.argument("revision", type=str)
def downgrade(revision):
    """Database Migrate (Downgrade)."""
    command.downgrade(ALEMBIC_CONFIG, revision)


@db.command()
def history():
    """Database Migrations History."""
    command.history(ALEMBIC_CONFIG)


@db.command()
def reset():
    """Database Delete Migrations (All-Versions)."""
    dir_path = pathlib.Path("./migrations/versions")
    # Delete
    try:
        shutil.rmtree(dir_path)
    except OSError as e:
        print(f"Error: {dir_path} : {e.strerror}")
    # Recreate
    os.makedirs(dir_path)
    with open(dir_path / "README", "w") as f:
        f.write("Database Migrations Versions")



"""
    <Regular> to <Async>
"""

import asyncio
from functools import wraps


def coro(f):
    """Transform <Regular> Function to <Async> Function."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        """Wraps the Function"""
        return asyncio.run(f(*args, **kwargs))

    return wrapper



"""
    Start A New Fastberry Project
"""

import os
import pathlib

import click

from .shell import shell_print, unzip

TEMPLATES_DIR = pathlib.Path(__file__).parents[0] / "templates"


@click.command()
def cli():
    """Fastberry Start-Project."""

    shell_print("""* Starting-Project! ...\n""")
    unzip(TEMPLATES_DIR / "project-template.zip", pathlib.Path(os.getcwd()))



""" [Plugin-Maker]
    Create Full-Plugin (Middleware, Extension, Permission, Router).
"""

MIDDLEWARE = '''
# -*- coding: utf-8 -*-
""" 
    [Middleware]
"""

from fastberry import BaseMiddleware

# Create <Middleware> here.

'''

EXTENSION = '''
# -*- coding: utf-8 -*-
""" 
    [Extension]
"""

from fastberry import BaseExtension

# Create <Extension> here.

'''

PERMISSION = '''
# -*- coding: utf-8 -*-
""" 
    [Permission]
"""

import typing

from strawberry.types import Info

from fastberry import BasePermission

# Create <Permission> here.

'''

ROUTER = '''
# -*- coding: utf-8 -*-
""" 
    [Router]
"""

from fastapi import APIRouter

router = APIRouter()

# Create <Router> here.

'''


def create_plugin(the_dir):
    """Create A Plugin"""

    with open(the_dir / "middleware.py", "w", encoding="utf-8") as file:
        file.write(MIDDLEWARE)

    with open(the_dir / "extension.py", "w", encoding="utf-8") as file:
        file.write(EXTENSION)

    with open(the_dir / "permission.py", "w", encoding="utf-8") as file:
        file.write(PERMISSION)

    with open(the_dir / "router.py", "w", encoding="utf-8") as file:
        file.write(ROUTER)



"""
    Shell Functions
"""

import pathlib
import shlex
import shutil
import subprocess
import zipfile

import click

TEMPORARY_DIR = pathlib.Path(__file__).parents[0] / "tmp"


def shell_commands(list_of_commands):
    """Shell Commands"""
    processes = []
    for cmd in list_of_commands:
        __code = shlex.split(cmd)
        task = subprocess.Popen(__code, shell=True)
        processes.append(task)
    return [process.wait() for process in processes]


def shell_print(text: str, color: str = "green"):
    """Shell Print"""
    return click.secho(f"{ text }", fg=color, bold=True)


def path_delete(dir_path):
    """Delete Path"""
    try:
        shutil.rmtree(dir_path)
    except OSError as e:
        print(f"Error: {dir_path} : {e.strerror}")


def unzip_base(source, destination):
    """Unzip Base"""
    with zipfile.ZipFile(source, "r") as zip_ref:
        zip_ref.extractall(destination)


def unzip(source: str, destination: str):
    """Unzip Method"""
    file_name = source.name
    file_name = file_name.replace(".zip", "")
    unzip_base(source, TEMPORARY_DIR)
    shutil.move(TEMPORARY_DIR / file_name, destination)



"""
    * SQLAlchemy + Databases — Controller
"""

import functools
import math
from types import SimpleNamespace
from typing import Any

from .sql_where import Filters as SQLFilters
from .utils import sql_id_decode, to_obj

try:
    from sqlalchemy.sql.elements import BinaryExpression
except ImportError:
    BinaryExpression = None

try:
    from databases import Database
except ImportError:
    def Database(x): return SimpleNamespace(database_url=x)


def sql_response(
    data: Any = None,
    error: bool = False,
    error_message: str | None = None,
    count: int | None = None,
    pages: int | None = None,
):
    """Database Transaction Response"""
    return SimpleNamespace(
        data=data, error=error, error_message=error_message, count=count, pages=pages
    )


def clean_form(base: Any, cols: list, form: dict):
    """Clean User's Input

    Args:
        form (dict): User's Input.

    Returns:
        dict: Clean User's Input.
    """
    inputs = {}
    base_form = base(**form)
    for key, val in base_form.__dict__.items():
        if key in cols and key != "_id":
            inputs[key] = val
    return inputs


def clean_update_form(base: Any, cols: list, form: dict):
    clean = clean_form(base, cols, form)
    return {key: val for key, val in clean.items() if val and key in form.keys()}

# Testing


class SQLBase:
    """SQlAlchemy & Databases (Manager)

    -----------------------------------------------------------------------------------------------
    # Init
    -----------------------------------------------------------------------------------------------
    sql = SQLBase(database_url, custom_type)

    -----------------------------------------------------------------------------------------------
    # Form
    -----------------------------------------------------------------------------------------------
    - Clean-Inputs      : sql.form(dict)
    - Clean-Empty       : sql.form_update(dict)

    -----------------------------------------------------------------------------------------------
    # Read (Examples)
    -----------------------------------------------------------------------------------------------
    - Filter-By         : await sql.filter_by(id=1, name="spongebob")
    - Search-Columns    : await sql.search(["name", "title"], "bob")
    - Get-By(Single-Row): await sql.get_by(id=1, name="spongebob")
    - Custom-Querying   : await sql.find(
                                    sql.Q.where("id", "in", [1, 2, 3]),
                                    page=1, limit=100, sort_by='-id'
                                )

    -----------------------------------------------------------------------------------------------
    # C.U.D
    -----------------------------------------------------------------------------------------------
    - Create            : await sql.create(dict)
    - Update            : await sql.update(items: list[IDs], values: dict)
    - Delete            : await sql.delete(items: list[IDs])
    """

    def __init__(self, database_url, custom_type):
        """Start Manager"""
        self.database = Database(database_url)
        self.table = custom_type.objects
        self.Q = SQLFilters(custom_type.objects)
        self.form = functools.partial(
            clean_form, custom_type, custom_type.objects.columns.keys())
        self.form_update = functools.partial(
            clean_update_form, custom_type, custom_type.objects.columns.keys())

    async def all(
        self,
    ):
        """Get All-Rows from Database"""
        items = await self.database.fetch_all(self.Q.select())
        return items

    async def find(
        self,
        search: BinaryExpression = None,
        page: int | None = None,
        limit: int | None = None,
        sort_by: str | None = None,
    ):
        """Get Multiple-Rows from Database Table by <SQLAlchemy-BinaryExpression>"""
        try:
            query = self.Q.find(search, page=page,
                                limit=limit, sort_by=sort_by)
            items = await self.database.fetch_all(query.query)
            count = await self.database.fetch_val(query.count)
            _limit = limit or 1
            pages = int(math.ceil(count / _limit))
            return sql_response(data=to_obj(items, sql=True), count=count, pages=pages)
        except:
            return sql_response(data=[], count=0, pages=0)

    async def filter_by(
        self,
        search: dict | None = None,
        page: int | None = None,
        limit: int | None = None,
        sort_by: str | None = None,
    ):
        """Get Multiple-Rows from Database Table by <Keyword-Arguments>"""
        query = self.Q.filter_by(**search) if search else None
        items = await self.find(query, page=page, limit=limit, sort_by=sort_by)
        return items

    async def search(
        self,
        columns: list | None = None,
        value: str | None = None,
        page: int | None = None,
        limit: int | None = None,
        sort_by: str | None = None,
    ):
        """Get Multiple-Rows from Database Table by <Searching-Columns>"""
        query = self.Q.search(columns, value) if value else None
        items = await self.find(query, page=page, limit=limit, sort_by=sort_by)
        return items

    async def get_by(self, **kwargs):
        """Get Single-Row from Database Table by <Keyword-Arguments>"""
        query = self.Q.filter_by(**kwargs)
        item = await self.database.fetch_one(self.Q.select(query))
        return to_obj(item, sql=True)

    async def detail(self, ID):
        """Get Single-Row from Database ID"""
        query = self.Q.filter_by(_id=sql_id_decode(ID))
        item = await self.database.fetch_one(self.Q.select(query))
        return to_obj(item, sql=True)

    async def create(self, form: dict):
        """Create Single-Row."""
        # Init Values
        return_value = sql_response()
        unique_id = False
        try:
            sql_query = self.table.insert(form)
            unique_id = await self.database.execute(sql_query)
            if unique_id:
                # If Success => Fetch Row
                return_value.data = await self.get_by(_id=unique_id)
        except Exception as error:
            return_value.error = True
            return_value.error_message = str(error)

        return return_value

    async def update(self, unique_ids: list[str], **kwargs):
        """Update Multiple/Single-Row(s)"""
        return_value = sql_response()
        # Get Ids
        all_ids = [sql_id_decode(i) for i in unique_ids]
        sql_ids_in = self.Q.where("_id", "in", all_ids)
        try:
            return_value.count = await self.database.execute(
                self.table.update().where(sql_ids_in).values(**kwargs)
            )
        except Exception as error:
            return_value.error = True
            return_value.error_message = str(error)
        # Get Details
        if len(all_ids) == 1 and return_value.count == 1:
            return_value.data = await self.get_by(_id=all_ids[0])
        return return_value

    async def delete(self, unique_ids: list[str]):
        """Delete Multiple/Single-Row(s)"""
        return_value = sql_response()
        # Get Ids
        all_ids = [sql_id_decode(i) for i in unique_ids]
        sql_ids_in = self.Q.where("_id", "in", all_ids)
        try:
            return_value.count = await self.database.execute(
                self.table.delete().where(sql_ids_in)
            )
        except Exception as error:
            return_value.error = True
            return_value.error_message = str(error)
        return return_value



""" Strawberry - GraphQL (Connection, PageInfo, Edges)
    Made Easy by wrapping functions together.

methods:
    Mutation: GraphQL Mutation.
    Query: GraphQL Query.
    Edges: GraphQL Edges.
    Response: Response GraphQL Edges.
    Error: Response Error([ErrorMessage]).
    ErrorMessage: Error Message.
"""

from dataclasses import field

import strawberry
from strawberry.scalars import JSON

from .edges import Connection, Edge, PageInfo


@strawberry.type
class ErrorMessage:
    """GraphQL Error Message"""

    type: str = None
    message: str = None


@strawberry.type
class Error:
    """GraphQL Error"""

    messages: list[ErrorMessage] = field(default_factory=list)
    meta: JSON = field(default_factory=dict)
    error: bool = True


def Query(model):
    """GraphQL Query"""
    return model | None


def Mutation(model):
    """GraphQL Mutation"""
    return model | Error


def Edges(model):
    """GraphQL Edges"""
    return Connection[model]


def Response(
    edges: list = None, length: int = None, pages: int = None, extra: dict = None
):
    """Edges Response"""
    edges = edges or []
    extra = extra or {}
    items = [Edge(node=item, cursor=item.id) for item in edges]
    return Connection(
        page_info=PageInfo(
            length=length,
            pages=pages,
            extra=extra,
        ),
        edges=items,
    )
